from pyFAI import azimuthalIntegrator
# import numpy
#
# import logging
#
# logger = logging.getLogger("pyFAI.azimuthalIntegrator")
#
# class Detector(pyFAI.Detector):
#     __statevars = (
#         '_pixel1', '_pixel2', '_pixel_corners', '_binning', '_mask', '_mask_crc', '_maskfile', '_splineFile', '_dx',
#         '_dy', '_flatfield', '_flatfield_crc', '_darkcurrent', '_darkcurrent_crc', '_splineCache', 'shape')
#
#     def __getnewargs_ex__(self):
#         return ((self.pixel1, self.pixel2, self.splineFile, self.max_shape), {})
#
#     def __getstate__(self):
#         vars = self.__statevars
#         return tuple(getattr(self, var) for var in vars)
#
#     def __setstate__(self, state):
#         for statevar, varkey in zip(state, self.__statevars):
#             setattr(self, varkey, statevar)
#         self.engines = {}
#
#
# # monkey patch to correct auto-inversion of masks when 'numpy' is used
class AzimuthalIntegrator(azimuthalIntegrator.AzimuthalIntegrator):

    def __deepcopy__(self, memo=None):
        """deep copy
        :param memo: dict with modified objects
        :return: a deep copy of itself."""
        numerical = ["_dist", "_poni1", "_poni2", "_rot1", "_rot2", "_rot3",
                     "chiDiscAtPi", "_dssa_order", "_wavelength",
                     '_oversampling', '_correct_solid_angle_for_spline',
                     '_transmission_normal',
                     ]
        if memo is None:
            memo = {}
        new = self.__class__()
        memo[id(self)] = new
        new_det = self.detector.__deepcopy__(memo)
        new.detector = new_det

        for key in numerical:
            old_value = self.__getattribute__(key)
            memo[id(old_value)] = old_value
            new.__setattr__(key, old_value)
        new_param = [new._dist, new._poni1, new._poni2,
                     new._rot1, new._rot2, new._rot3]
        memo[id(self.param)] = new_param
        new.param = new_param
        cached = {}
        memo[id(self._cached_array)] = cached
        try:
            for key, old_value in self._cached_array.copy().items():
                if "copy" in dir(old_value):
                    new_value = old_value.copy()
                    memo[id(old_value)] = new_value
            new._cached_array = cached
        except Exception:
            print('huh?')
        return new

#     def create_mask(self, data, mask=None,
#                     dummy=None, delta_dummy=None, mode="normal"):
#         """
#         Combines various masks into another one.
#
#         @param data: input array of data
#         @type data: ndarray
#         @param mask: input mask (if none, self.mask is used)
#         @type mask: ndarray
#         @param dummy: value of dead pixels
#         @type dummy: float
#         @param delta_dumy: precision of dummy pixels
#         @type delta_dummy: float
#         @param mode: can be "normal" or "numpy" (inverted) or "where" applied to the mask
#         @type mode: str
#
#         @return: the new mask
#         @rtype: ndarray of bool
#
#         This method combine two masks (dynamic mask from *data &
#         dummy* and *mask*) to generate a new one with the 'or' binary
#         operation.  One can adjust the level, with the *dummy* and
#         the *delta_dummy* parameter, when you consider the *data*
#         values needs to be masked out.
#
#         This method can work in two different *mode*:
#
#             * "normal": False for valid pixels, True for bad pixels
#             * "numpy": True for valid pixels, false for others
#
#         This method tries to accomodate various types of masks (like
#         valid=0 & masked=-1, ...) and guesses if an input mask needs
#         to be inverted.
#         """
#         shape = data.shape
#         #       ^^^^   this is why data is mandatory !
#         if mask is None:
#             mask = self.mask
#         if mask is None:
#             mask = numpy.zeros(shape, dtype=bool)
#         elif mask.min() < 0 and mask.max() == 0:  # 0 is valid, <0 is invalid
#             mask = (mask < 0)
#         else:
#             mask = mask.astype(bool)
#         # if mask.sum(dtype=int) > mask.size // 2:                              # TERRRIBLEE!
#         #     logger.warning("Mask likely to be inverted as more"
#         #                    " than half pixel are masked !!!")
#         #     numpy.logical_not(mask, mask)
#         if (mask.shape != shape):
#             try:
#                 mask = mask[:shape[0], :shape[1]]
#             except Exception as error:  # IGNORE:W0703
#                 logger.error("Mask provided has wrong shape:"
#                              " expected: %s, got %s, error: %s" %
#                              (shape, mask.shape, error))
#                 mask = numpy.zeros(shape, dtype=bool)
#         if dummy is not None:
#             if delta_dummy is None:
#                 numpy.logical_or(mask, (data == dummy), mask)
#             else:
#                 numpy.logical_or(mask,
#                                  abs(data - dummy) <= delta_dummy,
#                                  mask)
#         if mode == "numpy":
#             numpy.logical_not(mask, mask)
#         elif mode == "where":
#             mask = numpy.where(numpy.logical_not(mask))
#         return mask
#
#     __statevars = (
#         '_cached_array', '_dssa', '_dssa_crc', '_dssa_order', '_oversampling', '_correct_solid_angle_for_spline',
#         '_cosa',
#         '_transmission_normal', '_transmission_corr', '_transmission_crc', 'detector')
#
#     def __getnewargs_ex__(self):
#         return ((self.dist, self.poni1, self.poni2, self.rot1, self.rot2, self.rot3, self.pixel1, self.pixel2,
#                  self.splineFile, self.detector, self.wavelength), {})
#
#     def __getstate__(self):
#         vars = self.__statevars
#         return tuple(getattr(self, var) for var in vars)
#
#     def __setstate__(self, state):
#         for statevar, varkey in zip(state, self.__statevars):
#             setattr(self, varkey, statevar)
#         self.engines = {}
#
#
azimuthalIntegrator.__dict__['AzimuthalIntegrator'] = AzimuthalIntegrator
#
#
# new_ALL_DETECTORS = {}
# new_ALL_DETECTORS['Detector'] = pyFAI.detectors.__dict__['Detector'] = pyFAI.__dict__['Detector'] = Detector
#
# from distributed.protocol.serialize import register_serialization
# import dill
# def dilldumps(x):
#     return {}, [dill.dumps(x)]
#
#
# def dillloads(header, frames):
#     return dill.loads(frames[0])
#
#
# register_serialization(pyFAI.AzimuthalIntegrator, dilldumps, dillloads)
# register_serialization(pyFAI.Detector, dilldumps, dillloads)
#
# for name, detector in pyFAI.detectors.ALL_DETECTORS.items():
#     if name != 'detector':
#         new_ALL_DETECTORS['name'] = type(detector.__name__, tuple(
#             base for base in detector.__bases__ if not base.__name__ == 'Detector') + (Detector,),
#                                          dict(detector.__dict__))
#
#     register_serialization(detector, dilldumps, dillloads)
#
# from pyFAI.calibrant import Calibrant
#
# register_serialization(Calibrant, dilldumps, dillloads)
from pyFAI import geometry
from pyFAI import detectors
from pyFAI import calibrant



from xicam.plugins.operationplugin import operation
import numpy as np
from pyFAI.ext.reconstruct import reconstruct


# Wraps the C implemented 'reconstruct' function to provide inspect metadata
def _reconstruct(ndarray_data: np.ndarray,
                 ndarray_mask: np.ndarray = None,
                 dummy: float = None,
                 delta_dummy: float = None):
    return reconstruct(ndarray_data, ndarray_mask, dummy, delta_dummy)


inpaint = operation(_reconstruct, name='Inpaint (pyFAI.ext.reconstruct)')
# TODO: add categories



import mimetypes
import os
import re
from datetime import datetime
from pathlib import Path

import dask
import dask.array as da
import fabio
import numpy as np
from bluesky_live.run_builder import RunBuilder
from xicam.SAXS.ontology import NXsas

mimetypes.add_type('application/x-edf', '.edf')


def parse_txt(path):
    if not os.path.isfile(path):
        return dict()

    with open(path, 'r') as f:
        lines = f.readlines()

    paras = dict()

    # The 7.3.3 txt format is messy, with keyless values, and extra whitespaces

    keylesslines = 0
    for line in lines:
        cells = list(filter(None, re.split('[=:]+', line)))

        key = cells[0].strip()

        if cells.__len__() == 2:
            cells[1] = cells[1].split('/')[0]
            paras[key] = cells[1].strip()
        elif cells.__len__() == 1:
            keylesslines += 1
            paras['Keyless value #' + str(keylesslines)] = key

    return paras


def edf_ingestor(paths):
    # No way to know if this is reflection or transmission, so everything is reflection?
    incidence_angle = None
    detector_rotation = None
    beamline_energy = None
    detector_translate_x = None
    detector_translate_y = None

    txt_path = paths[0].replace('.edf', '.txt')
    if os.path.isfile(txt_path):
        extra_metadata = parse_txt(txt_path)

        # read incident angle
        if "Sample Alpha Stage" in extra_metadata:
            incidence_angle = np.deg2rad(float(extra_metadata["Sample Alpha Stage"]))

        detector_rotation = 0  # Assume no rotation

    projections = [{'name': 'NXsas',
                    'version': '0.1.0',
                    'projection':
                        {NXsas.DATA_PROJECTION_KEY: {'type': 'linked',
                                                     'stream': 'primary',
                                                     'location': 'event',
                                                     'field': 'image'}},
                    'configuration': {'geometry_mode': 'reflection'}
                    }]

    if incidence_angle is not None:
        projections[0]['projection'][NXsas.INCIDENCE_ANGLE_PROJECTION_KEY] = {'type': 'static',
                                                                              'value': incidence_angle}
    if detector_rotation is not None:
        projections[0]['projection'][NXsas.AZIMUTHAL_ANGLE_PROJECTION_KEY] = {'type': 'static',
                                                                              'value': detector_rotation}
    if beamline_energy is not None:
        projections[0]['projection'][NXsas.ENERGY_PROJECTION_KEY] = {'type': 'static',
                                                                     'value': beamline_energy}

    if detector_translate_x is not None:
        projections[0]['projection'][NXsas.DETECTOR_TRANSLATION_X_PROJECTION_KEY] = {'type': 'static',
                                                                                     'value': detector_translate_x}
    if detector_translate_y is not None:
        projections[0]['projection'][NXsas.DETECTOR_TRANSLATION_Y_PROJECTION_KEY] = {'type': 'static',
                                                                                     'value': detector_translate_y}

    d = []
    t = []

    lazy_read = dask.delayed(lambda path: np.flipud(fabio.open(path).data), pure=True)
    for path in paths:
        with fabio.open(path) as file:
            d.append(da.from_delayed(lazy_read(path), dtype=file.dtype, shape=file.shape))
            t.append(datetime.strptime(file.header['Date'], '%a %b  %d %H:%M:%S %Y').timestamp())

    d = np.stack(d)

    metadata = {'projections': projections, 'sample_name': Path(paths[0]).name}
    data_keys = {'image': {'source': 'Beamline 7.3.3',
                           'dtype': 'array',
                           'shape': d.shape,
                           'dims': ('dim_0', 'dim_1')}}
    with RunBuilder(metadata=metadata) as builder:
        builder.add_stream("primary",
                           data_keys=data_keys,
                           )
        builder.add_data("primary",
                         # NOTE: Put data in list, since Runbuilder.add_stream expects
                         # a sequence number to add event_page
                         data={'image': d},
                         time=t)

    builder.get_run()
    yield from builder._cache


if __name__ == '__main__':
    list(edf_ingestor(["E:\\data\\YL1031\\YL1031__2m_00000.edf"]))



import uuid

import numpy as np
from pyFAI.detectors import ALL_DETECTORS
from scipy import constants

from xicam.SAXS.intents import SAXSImageIntent, GISAXSImageIntent
from xicam.SAXS.ontology import NXsas
from xicam.SAXS.patches.pyFAI import AzimuthalIntegrator
from xicam.core.data import ProjectionNotFound
from xicam.core import msg
from xicam.SAXS.ontology.NXsas import DATA_PROJECTION_KEY
from xicam.plugins import manager as plugin_manager

PROJECTION_NAME = "NXsas"
PROJECTION_KEYS = {
    "Detector Rotation": "",
    "Detector Translate": "",
    "detector_distance": "",
    "Beamline Energy": "",
    "Sample Translate": "",
    "Sample Rotation": "",
}

# Pinhole_X
# Pinhole_Y
# Sample Lift
# Sample Translate
# Sample Rotate Stepper (theta)
# Detector Rotate
# Det Translate
# Beamline Energy
# LS_LLH T A
# LS_LLH T B
# Mono energy,
# EPU energy


def extract_mapped_value(run_catalog, projection, key):
    # TODO: should safely handle if key not found in projection
    if projection['projection'][key]['type'] == 'linked':
        stream = projection['projection'][key]['stream']
        field = projection['projection'][key]['field']
        return getattr(run_catalog, stream).to_dask()[field]
    elif projection['projection'][key]['type'] == 'static':
        return projection['projection'][key]['value']


def project_NXsas(run_catalog):
    projection = next(
        filter(lambda projection: projection['name'] == PROJECTION_NAME, run_catalog.metadata['start']['projections']),
        None)

    if not projection:
        raise ProjectionNotFound(f"Could not find projection named '{PROJECTION_NAME}'.")

    data = extract_mapped_value(run_catalog, projection, NXsas.DATA_PROJECTION_KEY)

    # try to extract dark image
    device_name = projection['projection'][DATA_PROJECTION_KEY]['field']
    try:
        darks = run_catalog.dark.to_dask()[device_name]
    except (AttributeError, KeyError) as e:
        darks = None
        msg.logMessage(e, level=msg.WARNING)

    # handle case where we don't have info to construct a geometry
    incidence_angle = None

    calibration_settings = plugin_manager.get_plugin_by_name('xicam.SAXS.calibration', 'SettingsPlugin')

    try:
        incidence_angle = extract_mapped_value(run_catalog, projection, NXsas.INCIDENCE_ANGLE_PROJECTION_KEY)

        detector_rotation = extract_mapped_value(run_catalog, projection, NXsas.AZIMUTHAL_ANGLE_PROJECTION_KEY)

        beamline_energy = extract_mapped_value(run_catalog, projection, NXsas.ENERGY_PROJECTION_KEY)

        detector_translate_x = extract_mapped_value(run_catalog, projection, NXsas.DETECTOR_TRANSLATION_X_PROJECTION_KEY)
        detector_translate_y = extract_mapped_value(run_catalog, projection, NXsas.DETECTOR_TRANSLATION_Y_PROJECTION_KEY)

        wavelength = 1.239842e-6 / beamline_energy  # convert from eV to meters

        # TODO: handle dynamic poni values
        poni1 = projection['configuration']['poni1'] - detector_translate_x
        poni2 = projection['configuration']['poni2'] - detector_translate_y

        sdd = projection['configuration']['sdd']

        # These are static values, take first one; using max() to get scalar value
        poni1 = poni1[0].values.max()
        rot2 = detector_rotation[0].values.max()
        wavelength = wavelength[0].values.max()
        incidence_angle = incidence_angle[0].values.max()

        # Create detector from projection metadata
        detector_name = projection['configuration']['detector_name']
        detector_class = ALL_DETECTORS[detector_name]
        detector = detector_class()

        # Convert poni from pixel to meters
        poni1 *= detector.get_pixel1()
        poni2 *= detector.get_pixel2()

        geometry = AzimuthalIntegrator(dist=sdd,
                                       poni1=poni1,
                                       poni2=poni2,
                                       rot2=-np.radians(rot2),  # Convert to radians, account for upward rotation
                                       detector=detector,
                                       wavelength=wavelength)

        calibration_settings.setAI(geometry, device_name)

    except (AttributeError, KeyError) as e:
        geometry = None
        msg.logMessage(e, level=msg.WARNING)

    if geometry is None and calibration_settings.child(device_name):
        geometry = calibration_settings.AI(device_name)

    intents_list = []
    if projection['configuration'].get('geometry_mode') == 'reflection':
        intents_list.append(GISAXSImageIntent(image=data,
                                              darks=darks,
                                              name=f"GISAXS 〈{run_catalog.metadata['start']['sample_name']}〉",
                                              geometry=geometry,
                                              incidence_angle=incidence_angle,
                                              match_key=uuid.uuid4(),
                                              device_name=device_name), )
    else:
        intents_list.append(SAXSImageIntent(image=data,
                                            darks=darks,
                                            name=f"SAXS 〈{run_catalog.metadata['start']['sample_name']}〉",
                                            geometry=geometry,
                                            match_key=uuid.uuid4(),
                                            device_name=device_name))

    return intents_list



from xicam.core.intents import PlotIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np
from typing import Tuple
from pyFAI import units
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from typing import Union


@operation
@display_name('q Integration')
@output_names('q', 'I')
@describe_input('integrator', 'A PyFAI.AzimuthalIntegrator object')
@describe_input('data', '2d array representing intensity for each pixel')
@describe_input('npt', 'Number of bins along q')
@describe_input('polz_factor', 'Polarization factor for correction')
@describe_input('unit', 'Output units for q')
@describe_input('radial_range', 'The lower and upper range of the radial unit. If not provided, range is simply '
                                '(data.min(), data.max()). Values outside the range are ignored.')
@describe_input('azimuth_range', 'The lower and upper range of the azimuthal angle in degree. If not provided, '
                                 'range is simply (data.min(), data.max()). Values outside the range are ignored.')
@describe_input('mask', 'Array (same size as image) with 1 for masked pixels, and 0 for valid pixels')
@describe_input('dark', 'Dark noise image')
@describe_input('flat', 'Flat field image')
@describe_input('method', 'Can be "numpy", "cython", "BBox" or "splitpixel", "lut", "csr", "nosplit_csr", '
                          '"full_csr", "lut_ocl" and "csr_ocl" if you want to go on GPU. To Specify the device: '
                          '"csr_ocl_1,2"')
@describe_input('normalization_factor', 'Value of a normalization monitor')
@describe_output('q2', 'Q squared bin center positions')
@describe_output('ln_I', 'Logarithm of Binned/pixel-split integrated intensity')
@intent(PlotIntent, name="Guinier plot", output_map={"x": "q^2", "y": "ln_I"}, labels={"bottom": "q^2",
                                                                                       "left": "ln_I"})
@categories(("Scattering", "Integration"))
def guinier_plot(integrator: AzimuthalIntegrator,
                data: np.ndarray,
                npt: int = 1000,
                polz_factor: float = 0,
                unit: Union[str, units.Unit] = "q_A^-1",
                radial_range: Tuple[float, float] = None,
                azimuth_range: Tuple[float, float] = None,
                mask: np.ndarray = None,
                dark: np.ndarray = None,
                flat: np.ndarray = None,
                method: str = 'splitbbox',
                normalization_factor: float = 1, ) -> Tuple[np.ndarray, np.ndarray]:
    q, I = integrator.integrate1d(data=data,
                                  npt=npt,
                                  radial_range=radial_range,
                                  azimuth_range=azimuth_range,
                                  mask=mask,
                                  polarization_factor=polz_factor,
                                  dark=dark,
                                  flat=flat,
                                  method=method,
                                  unit=unit,
                                  normalization_factor=normalization_factor)
    q2 = q**2
    ln_I = np.log(I)
    return q2, ln_I



from astropy.modeling.models import Gaussian1D, Shift
from xicam.plugins.fittablemodelplugin import Fittable1DModelPlugin


class ShiftedGaussian1D(Fittable1DModelPlugin):
    def __new__(cls, *args, **kwargs):
        return Gaussian1D(*args, **kwargs) + Shift(*args, **kwargs)



import numpy as np
import skbeam.core.correlation as corr
from astropy.modeling import Fittable1DModel, Parameter, fitting
from xicam.core.intents import PlotIntent

from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent, visible
from typing import Union, Tuple


@operation
@display_name('Fit Scattering Factor')
@output_names('fit_curve', 'relaxation_rates', "tau", "g2")
@describe_input('g2', 'Normalized intensity-intensity time autocorrelation')
@describe_input('tau', 'delay time')
@describe_input('beta', 'Optical contrast (speckle contrast), a sample-independent beamline parameter')
@describe_input('baseline', 'baseline of one time correlation equal to one for ergodic samples')
@describe_input('correlation_threshold', 'threshold defining which g2 values to fit')
@describe_output('fit_curve', 'Fitted model of the g2 curve')
@describe_output('relaxation_rates', 'Relaxation time associated with the samples dynamics')
@intent(PlotIntent,
        canvas_name="1-time Correlation",
        match_key='1-time Correlation',
        name='g2 fit',
        xLogMode=True,
        labels={"bottom": "𝜏", "left": "g₂"},
        output_map={'x': 'tau', 'y': 'fit_curve'},
        mixins=["ToggleSymbols"])
@intent(PlotIntent,
        canvas_name="1-time Correlation",
        match_key='1-time Correlation',
        name='g2',
        xLogMode=True,
        labels={"bottom": "𝜏", "left": "g₂"},
        output_map={'x': 'tau', 'y': 'g2'},
        mixins=["ToggleSymbols"])
@categories(('Scattering', 'Fitting'))
@visible('g2', False)
@visible('tau', False)
def fit_scattering_factor(g2: np.ndarray,
                          tau: np.ndarray,
                          beta: float = 1.0,
                          baseline: float = 1.0,
                          relaxation_rate: float = 0.01,
                          correlation_threshold: float = 2) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    model = ScatteringModel(beta, baseline, relaxation_rate=relaxation_rate)
    fitting_algorithm = fitting.SLSQPLSQFitter()
    threshold = min(len(tau), np.argmax(g2 < correlation_threshold))

    if g2.ndim > 1:
        fits = [fitting_algorithm(model, tau[:threshold], g2[i][:threshold]) for i in range(len(g2))]
    else:
        fits = [fitting_algorithm(model, tau[:threshold], g2[:threshold])]

    relaxation_rates = np.asarray([fit.relaxation_rate.value for fit in fits]).squeeze()
    fit_curves = np.asarray([fit(tau) for fit in fits]).squeeze()

    return fit_curves, relaxation_rates, tau, g2

    # labels = {'left': ['g<sub>2</sub>(&tau;)', 's'],
    #             'bottom': ['&tau;', 's']}
    # one_time_hint = PlotHint(self.lag_steps.value[1:], self.g2.value[1:], name="1-Time", labels=labels, xLog=True, style=Qt.SolidLine)
    # fit_hint = PlotHint(self.lag_steps.value[1:], self.fit_curve.value[1:], name="1-Time Fit", labels=labels, xLog=True, style=Qt.DashLine)
    # self.intents = [CoPlotHint(one_time_hint, fit_hint, name="1-Time")]


class ScatteringModel(Fittable1DModel):
    inputs = ('tau',)
    outputs = ('g2',)

    relaxation_rate = Parameter()

    def __init__(self, beta, baseline=1.0, **kwargs):
        self.beta = beta
        self.baseline = baseline
        super(ScatteringModel, self).__init__(**kwargs)

    def evaluate(self, tau, relaxation_rate):
        return corr.auto_corr_scat_factor(tau, self.beta, relaxation_rate, self.baseline)

    def fit_deriv(self, tau, relaxation_rate):
        d_relaxation_rate = -2 * self.beta * relaxation_rate * np.exp(-2 * relaxation_rate * tau)
        return [d_relaxation_rate]



from xicam.SAXS.widgets.SAXSToolbar import SAXSToolbarBase, ROIs
from xicam.plugins import manager as plugin_manager
from xicam.plugins import live_plugin
from xicam.gui.canvases import ImageIntentCanvas
from xicam.gui.widgets.imageviewmixins import LogScaleIntensity, ImageViewHistogramOverflowFix, \
    QCoordinates, Crosshair, BetterButtons, CenterMarker, ToolbarLayout, EwaldCorrected, ROICreator

from xicam.SAXS.widgets.imageviewmixins import BackgroundCorrected


@live_plugin("ImageMixinPlugin")
class SAXSToolbarMixin(ToolbarLayout):
    def __init__(self, *args, **kwargs):
        # Toolbar api will need view function or object...
        toolbar = ROIs()
        super(SAXSToolbarMixin, self).__init__(*args, toolbar=toolbar, **kwargs)


# FIXME: investigate EWaldCorrected (particularly ProcessingView) -- causes "tuple indexing error" on ProxyView when opening image
@live_plugin("ImageMixinPlugin")
class SAXSImageIntentBlend(BackgroundCorrected, LogScaleIntensity, CenterMarker, BetterButtons, Crosshair, QCoordinates,
                           ImageViewHistogramOverflowFix, ROICreator, EwaldCorrected):  # LogButtons):
    ...


class SAXSImageIntentCanvas(ImageIntentCanvas):
    def __init__(self, *args, **kwargs):
        super(SAXSImageIntentCanvas, self).__init__(*args, **kwargs)

    def render(self, intent, **_):
        super(SAXSImageIntentCanvas, self).render(intent, mixins=["SAXSImageIntentBlend"])
        if hasattr(intent, "geometry"):
            self.canvas_widget.setGeometry(intent.geometry)
        if hasattr(intent, "darks") and hasattr(self.canvas_widget, "set_darks"):
            self.canvas_widget.set_darks(intent._darks)
        if hasattr(intent, "log_scale"):
            self.canvas_widget.setLogScale(intent.log_scale)



from xicam.core.execution.workflow import Workflow
from .detector import detector_mask_plugin


class MaskingWorkflow(Workflow):
    def __init__(self):
        super(MaskingWorkflow, self).__init__('Masking')

        detectormask = detector_mask_plugin()

        self.add_operation(detectormask)
        self.auto_connect_all()



from xicam.core.intents import ImageIntent


class SAXSImageIntent(ImageIntent):

    canvas = "saxs_image_intent_canvas"

    def __init__(self, name: str, image, *args, device_name: str = None, geometry=None, log_scale=True, **kwargs):
        super(SAXSImageIntent, self).__init__(name, image, *args, **kwargs)

        self.geometry = geometry
        self.log_scale = log_scale
        self.device_name = device_name


class GISAXSImageIntent(SAXSImageIntent):

    def __init__(self, name: str, image, *args, incidence_angle=0.0, **kwargs):
        super(GISAXSImageIntent, self).__init__(name, image, *args, **kwargs)

        self.incidence_angle = incidence_angle



def test_onetime(self):
    # catalog.upsert(start, stop, doc_gen, [], {})
    data = np.random.random((100, 10, 10))
    labels = np.ones(data.shape[1:])
    print(data.shape)
    print(labels.shape)
    op = onetime.one_time_correlation()
    workflow = Workflow(operations=(op,))
    result = workflow.execute_synchronous(data=data, labels=labels)
    print(result)


from xicam.SAXS.intents import SAXSImageIntent
from xicam.core.data import ProjectionNotFound
from xicam.core.data.bluesky_utils import display_name
from xicam.core.intents import ImageIntent
from xicam.SAXS.ingestors.nxcansas import I_PROJECTION_KEY, QX_PROJECTION_KEY, QY_PROJECTION_KEY


# TODO: Intent for experimental geometry (SAXSImageIntent will include the geometry)
#  allow NOT having a geometry as well
def project_nxcanSAS(run_catalog):
    projection = next(
        filter(lambda projection: projection['name'] == 'NXcanSAS', run_catalog.metadata['start'].get('projections', [])), None)
    if not projection:
        raise ProjectionNotFound("Could not find projection 'NXcanSAS'.")
    catalog_name = display_name(run_catalog).split(" ")[0]

    data_stream = projection['projection'][I_PROJECTION_KEY]['stream']
    data_field = projection['projection'][I_PROJECTION_KEY]['field']
    qx_stream = projection['projection'][QX_PROJECTION_KEY]['stream']
    qx_field = projection['projection'][QX_PROJECTION_KEY]['field']
    qy_stream = projection['projection'][QY_PROJECTION_KEY]['stream']
    qy_field = projection['projection'][QY_PROJECTION_KEY]['field']

    data = getattr(run_catalog, data_stream).to_dask().rename({data_field: I_PROJECTION_KEY})
    qx = getattr(run_catalog, data_stream).to_dask().rename({qx_field: QX_PROJECTION_KEY})
    qy = getattr(run_catalog, data_stream).to_dask().rename({qy_field: QY_PROJECTION_KEY})

    intents_list = []
    # TODO: construct geometry from qx / qy, then pass into intent below
    intents_list.append(SAXSImageIntent(image=data, item_name='Intensity'))

    return intents_list


import numpy as np
from xicam.plugins.operationplugin import operation, output_names, visible
from xicam.plugins import live_plugin

@live_plugin("OperationPlugin")
@operation
@output_names("images")
@visible("images", False)
def synthetic_image_series(images: np.ndarray, n: int = 100) -> np.ndarray:
    # return np.array([ione_timemages for i in range(n)])
    return np.array([np.random.poisson(np.where(images >= 0, images, 0)) for i in range(n)])



from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
from pyFAI import calibrant
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import numpy as np


@operation
@display_name('Simulate Calibrant')
@output_names('data')
@describe_output('data', 'Simulated calibrant image data')
@describe_input('calibrant', "Calibrant standard record")
@describe_input('azimuthal_integrator', "Azimuthal integrator; the SDD will be modified in-place")
@describe_input('I_max', 'Maximum intensity of rings')
@categories(('Scattering', 'Calibration'))
def simulate_calibrant(azimuthal_integrator: AzimuthalIntegrator,
                       calibrant: calibrant.Calibrant,
                       I_max: float = 1) -> np.ndarray:
    calibrant.set_wavelength(azimuthal_integrator.get_wavelength())
    data = calibrant.fake_calibration_image(azimuthal_integrator, Imax=I_max)

    return data



from xicam.SAXS.intents import SAXSImageIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np


@operation
@display_name('Chi squared')
@output_names('chi_squared')
@describe_input('dataA', 'Frame A image data')
@describe_input('dataB', 'Frame B image data')
@describe_output('chi_squared', 'Chi-squared difference between consecutive frames')
@intent(SAXSImageIntent, name='Chi squared', output_map={'image': 'chi_squared'})
@categories(('General', 'Mathematics'))
def chi_squared(dataA: np.ndarray,
                dataB: np.ndarray) -> np.ndarray:
    chi_squared = (dataA - dataB) ** 2.

    return chi_squared



from xicam.plugins import OperationPlugin
import numpy as np
from astropy.modeling import fitting
from astropy.modeling import Fittable1DModel
from typing import Tuple
from enum import Enum
from xicam.plugins import manager as pluginmanager
from pyqtgraph.parametertree import Parameter


class AstropyQSpectraFit(OperationPlugin):
    name = 'Q Fit (Astropy)'

    def __init__(self):
        super(AstropyQSpectraFit, self).__init__()

    def as_parameter(self):
        return [{"name": "Model", "type": "list", "limits": self.limits.get('model', {plugin.name: plugin for plugin in
                                                                                      pluginmanager.get_plugins_of_type(
                                                                                          'Fittable1DModelPlugin')})},
                {"name": "Model Parameters", "type": "group"}]

    def wireup_parameter(self, parameter):
        parameter.sigValueChanged.connect(self.value_changed)

    def value_changed(self, *args, **kwargs):
        print(args, kwargs)

    # TODO: model parameters could be set in the gui, but can they be connected programmatically?
    def _func(self, q, I, model: Fittable1DModel, domain_min=-np.inf, domain_max=np.inf, fitter=fitting.LevMarLSQFitter,
              **model_parameters):
        for name, value in model_parameters.items():  # propogate user-defined values to the model
            getattr(model, name).value = value
            getattr(model, name).fixed = self.fixed.get(name)

        filter = np.logical_and(domain_min <= q, q <= domain_max)
        q = q[filter]
        I = I[filter]
        fitted_model = fitter()(model, q, I)
        fitted_profile = fitted_model(q)

        return fitted_model, fitted_profile

#
# class AstropyQSpectraFit(ProcessingPlugin):
#     name = 'Q Fit (Astropy)'
#
#     q = InOut(description='Q bin center positions',
#               type=np.array)
#     Iq = InOut(description='Q spectra bin intensities', type=np.array)
#     model = Input(description='Fittable model class in the style of Astropy', type=Enum)
#     domainmin = Input(description='Min bound on the domain of the input data', type=float)
#     domainmax = Input(description='Max bound on the domain of the input data', type=float)
#     fitter = Input(description='Fitting algorithm', default=fitting.LevMarLSQFitter(), type=Enum,
#                    limits={'Linear LSQ': fitting.LinearLSQFitter(),
#                            'Levenberg-Marquardt LSQ': fitting.LevMarLSQFitter(), 'SLSQP LSQ': fitting.SLSQPLSQFitter(),
#                            'Simplex LSQ': fitting.SimplexLSQFitter()})
#
#     fittedmodel = Output(description='A new model with the fitted parameters; behaves as parameterized function',
#                          type=Fittable1DModel)
#     fittedprofile = Output(
#         description='The fitted profile from the evaluation of the resulting model over the input range.')
#
#     intents = [PlotHint(q, Iq), PlotHint(q, fittedprofile)]
#
#     modelvars = {}
#
#     def __init__(self):
#         super(AstropyQSpectraFit, self).__init__()
#         self.model.limits = {plugin.name: plugin for plugin in
#                              pluginmanager.get_plugins_of_type('Fittable1DModelPlugin')}
#         self.model.value = list(self.model.limits.values())[0]
#
#     @property
#     def parameter(self):
#
#         # clear cache in
#         for input in self.modelvars:
#             del self.__dict__[input]
#         varcache = self.modelvars.copy()
#         self.modelvars = {}
#         self._inputs = None
#         self._inverted_vars = None
#         if hasattr(self, '_inverted_vars'): del self._inverted_vars
#
#         for name in self.model.value.param_names:
#             param = getattr(self.model.value, name)
#             # TODO: CHECK NAMESPACE
#             if name in varcache:
#                 input = varcache[name]
#             else:
#                 input = InOut(name=name, default=param.default, limits=param.bounds, type=float, fixed=False,
#                               fixable=True)
#             setattr(self, name, input)
#             self.modelvars[name] = input
#         parameter = super(AstropyQSpectraFit, self).parameter
#         parameter.child('model').sigValueChanged.connect(self.reset_parameter)
#         return parameter
#
#     def wire
#
#     def reset_parameter(self):
#         # cache old parameter
#         oldparam = self._param
#
#         # empty it
#         for child in oldparam.children():
#             child.remove()
#
#         # reset attribute so new parameter is generated
#         self._param = None
#
#         # add new children to old parameter
#         for child in self.parameter.children():  # type: Parameter
#             oldparam.addChild(child)
#
#         # set old parameter to attribute
#         self._param = oldparam
#
#     def evaluate(self):
#         if self.model.value is None or self.model.value == '----': return
#         norange = self.domainmin.value == self.domainmax.value
#         if self.domainmin.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
#             self.domainmin.value = self.q.value.min()
#         if self.domainmax.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
#             self.domainmax.value = self.q.value.max()
#         for name, input in self.modelvars.items():  # propogate user-defined values to the model
#             getattr(self.model.value, name).value = input.value
#             getattr(self.model.value, name).fixed = input.fixed
#         filter = np.logical_and(self.domainmin.value <= self.q.value, self.q.value <= self.domainmax.value)
#         q = self.q.value[filter]
#         Iq = self.Iq.value[filter]
#         self.fittedmodel.value = self.fitter.value(self.model.value, q, Iq)
#         self.fittedprofile.value = self.fittedmodel.value(self.q.value)



from qtpy.QtGui import *
from qtpy.QtCore import Signal, Qt
from qtpy.QtWidgets import *
from xicam.gui.static import path
from pyFAI import detectors
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from pyFAI.multi_geometry import MultiGeometry

from xicam.plugins import ParameterSettingsPlugin
from xicam.core import msg
from .CalibrationPanel import CalibrationPanel

from pyqtgraph.parametertree import Parameter, ParameterTree
from pyqtgraph.parametertree.parameterTypes import GroupParameter, ListParameter, SimpleParameter
from xicam.SAXS.detectors import FastCCD


# https://stackoverflow.com/questions/20866996/how-to-compress-slot-calls-when-using-queued-connection-in-qt

# TODO: Refactor this class to be a view on the AI
class DeviceParameter(GroupParameter):
    def __init__(self, device, **opts):
        opts['type'] = 'bool'
        opts['value'] = True
        opts['name'] = device
        opts['removable'] = True
        ALL_DETECTORS = {(getattr(detector, 'aliases', ) or [detector.__name__])[0]: detector for detector in
                         detectors.ALL_DETECTORS.values()}
        geometrystyle = ListParameter(name='Geometry Style', type='list',
                                      values=['Fit2D', 'pyFAI', 'wxDiff'], value='Fit2D')
        detector = ListParameter(name='Detector', type='list', values=ALL_DETECTORS, value=ALL_DETECTORS['Fast CCD'])
        pixelx = SimpleParameter(name='Pixel Size X', type='float', value=172.e-6, siPrefix=True, suffix='m')
        pixely = SimpleParameter(name='Pixel Size Y', type='float', value=172.e-6, siPrefix=True, suffix='m')
        binning = SimpleParameter(name='Binning', type='int', value=1, suffix='x', limits=(1, 100))
        centerx = SimpleParameter(name='Center X', type='float', value=0, suffix=' px', decimals=4)
        centery = SimpleParameter(name='Center Y', type='float', value=0, suffix=' px', decimals=4)
        sdd = SimpleParameter(name='Detector Distance', type='float', value=1, siPrefix=True, suffix='m',
                              limits=(0, 1000), step=1e-3)
        tilt = SimpleParameter(name='Detector Tilt', type='float', value=0, siPrefix=False, suffix=u'°')
        rotation = SimpleParameter(name='Detector Rotation', type='float', value=0, siPrefix=False, suffix=u'°')

        self.children = [geometrystyle, detector, pixelx, pixely, binning, centerx, centery, sdd, tilt, rotation]
        opts['children'] = self.children
        super(DeviceParameter, self).__init__(**opts)


class DeviceProfiles(ParameterSettingsPlugin):
    sigGeometryChanged = Signal(AzimuthalIntegrator)  # Emits the new geometry

    def __init__(self):
        self.headermodel = None
        self.selectionmodel = None
        self.multiAI = MultiGeometry([])
        self.AIs = dict()
        self._changes = []
        self.isSilent = False


        energy = SimpleParameter(name='Energy', type='float', value=10000, siPrefix=True, suffix='eV')
        wavelength = SimpleParameter(name='Wavelength', type='float', value=1.239842e-6 / 10000, siPrefix=True,
                                     suffix='m')

        icon = QIcon(str(path('icons/calibrate.png')))
        super(DeviceProfiles, self).__init__(icon, "Device Profiles", [energy, wavelength], addText='New Device')

        self.sigTreeStateChanged.connect(self.stateChanged)

    def normalize(self, changes):
        for parameter, key, value in changes:
            if parameter.name == 'Wavelength':
                self.param('Energy').setValue(1.239842e-6 / self.param('Wavelength').value(),
                                              blockSignal=self.parameter.sigTreeStateChanged)
            elif parameter.name == 'Energy':
                self.param('Wavelength').setValue(1.239842e-6 / self.param('Energy').value(),
                                                  blockSignal=self.WavelengthChanged)

    def stateChanged(self, parent, changes):
        self.normalize(changes)

        self._changes.extend(changes)
        if not self.isSilent:
            self.emitChanges()
            self.apply()
            self.save()

    def emitChanges(self):
        modified_AIs = {self.AI(change[0].parent().name()) for change in self._changes if change[0].parent()}
        self._changes = []

        self.genAIs()
        for ai in modified_AIs:
            self.sigGeometryChanged.emit(ai)

    def setSilence(self, silence):
        self.isSilent = silence

        if not silence:
            self.emitChanges()
            self.apply()
            self.save()

    def addNew(self, typ=None):
        text, ok = QInputDialog().getText(self.widget, 'Enter Device Name', 'Device Name:')

        if text and ok:
            self.addDevice(text)

    def genAIs(self):

        for parameter in self.children():
            if isinstance(parameter, DeviceParameter):
                device = parameter.name()
                ai = self.AI(device)
                ai.set_wavelength(self['Wavelength'])
                ai.detector = parameter['Detector']()
                ai.detector.set_binning([parameter['Binning']] * 2)
                ai.detector.set_pixel1(parameter['Pixel Size X'])
                ai.detector.set_pixel2(parameter['Pixel Size Y'])
                fit2d = ai.getFit2D()
                fit2d['centerX'] = parameter['Center X']
                fit2d['centerY'] = ai.detector.shape[0] - parameter['Center Y']
                fit2d['directDist'] = parameter['Detector Distance'] * 1000
                fit2d['tilt'] = parameter['Detector Tilt']
                fit2d['tiltPlanRotation'] = parameter['Detector Rotation']
                ai.setFit2D(**fit2d)

    def AI(self, device):
        if device not in self.AIs:
            self.addDevice(device)
        return self.AIs.get(device, None)

    def setAI(self, ai: AzimuthalIntegrator, device: str):
        if not self.child(device):
            self.addDevice(device)

        self.AIs[device] = ai
        self.multiAI.ais = self.AIs.values()

        # propagate new ai to parameter
        fit2d = ai.getFit2D()
        try:
            self.setSilence(True)
            self.child(device, 'Detector').setValue(type(ai.detector))
            self.child(device, 'Binning').setValue(ai.detector.binning[0])
            self.child(device, 'Detector Tilt').setValue(fit2d['tiltPlanRotation'])
            self.child(device, 'Detector Rotation').setValue(fit2d['tilt'])
            self.child(device, 'Pixel Size X').setValue(ai.pixel1)
            self.child(device, 'Pixel Size Y').setValue(ai.pixel2)
            self.child(device, 'Center X').setValue(fit2d['centerX'])
            self.child(device, 'Center Y').setValue(ai.detector.shape[0] - fit2d['centerY'])
            self.child(device, 'Detector Distance').setValue(fit2d['directDist'] / 1000.)
            self.child('Wavelength').setValue(ai.wavelength)
        finally:
            self.setSilence(False)



    def addDevice(self, device):
        if device:
            try:
                self.setSilence(True)
                devicechild = DeviceParameter(device)
                self.addChild(devicechild)
                ai = AzimuthalIntegrator(wavelength=self['Wavelength'])
                ai.detector = FastCCD()
                self.AIs[device] = ai
                self.multiAI.ais = list(self.AIs.values())
            finally:
                self.setSilence(False)

    def setModels(self, headermodel, selectionmodel):
        self.headermodel = headermodel
        self.headermodel.dataChanged.connect(self.dataChanged)
        self.selectionmodel = selectionmodel


    def dataChanged(self, start, end):
        currentIndex = self.selectionmodel.currentIndex()
        if currentIndex.isValid():
            # TODO-- remove hard-coding of stream
            stream = "primary"
            item = self.headermodel.item(self.selectionmodel.currentIndex().row())
            catalog = item.data(Qt.UserRole)  # type: Catalog
            fields = [technique["data_mapping"]["data_image"][1] for technique in catalog.metadata["techniques"] if
                      technique["technique"] == "scattering"]

            for field in fields:
                if field not in self.AIs:
                    self.addDevice(field)

    def wavelengthChanged(self):
        self.param('Energy').setValue(1.239842e-6 / self.param('Wavelength').value(), blockSignal=self.EnergyChanged)

    def energyChanged(self):
        self.param('Wavelength').setValue(1.239842e-6 / self.param('Energy').value(),
                                          blockSignal=self.WavelengthChanged)

    def toState(self):
        self.apply()
        return self.saveState(filter='user'), self.AIs

    def fromState(self, state):
        self.setSilence(True)
        try:
            self.AIs = state[1]
            for child in self.children()[2:]:
                child.remove()
            for name, ai in self.AIs.items():
                if name in state[0]['children']:
                    self.addDevice(name)
                    self.setAI(ai, name)

            self.restoreState(state[0], addChildren=False, removeChildren=False)
        except Exception as ex:
            msg.logError(ex)
        finally:
            self.setSilence(False)



rom
xicam.plugins.operationplugin
import operation, output_names, \
    display_name, describe_input, \
    describe_output, categories
import numpy as np


# TODO: Delete horizontalcuts.py, seems unused.
@operation
@display_name("Horizontal Cut")
@output_names('horizontal_cut')
@describe_input('data', 'Frame image data')
@describe_input('q_x', 'horizontal q coordinate corresponding to data')
@describe_input('mask', 'Frame image data')
@describe_input('q_x_min', 'horizontal q minimum limit')
@describe_input('q_x_max', 'horizontal q maximum limit')
@describe_output('horizontal_cut', 'mask (1 is masked) with dimension of data')
# TODO: add categories

def horizontal_cut(data: np.ndarray,
                   q_x: np.ndarray,
                   mask: np.ndarray = None,
                   q_x_min: int,
                   q_x_max: int) -> np.ndarray:
    if mask is not None:
        horizontal_cut = np.logical_or(mask, q_x < q_x_min, q_x > q_x_max)
    else:
        horizontal_cut = np.logical_or(q_x < q_x_min, q_x > q_x_max)
    return horizontal_cut



"""
Usage: pip install -e .
       python setup.py install
       python setup.py bdist_wheel
       python setup.py sdist bdist_egg
       twine upload dist/*
"""
from os import path
from setuptools import find_namespace_packages, setup
import versioneer

here = path.abspath(path.dirname(__file__))
# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [x.strip() for x in all_reqs]  # if 'git+' not in x]

setup(
    name='xicam.SAXS',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="SAXS GUI Interface",

    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://github.com/ronpandolfi/Xi-cam',

    # Author details
    author='Ronald J Pandolfi',
    author_email='ronpandolfi@lbl.gov',

    # Choose your license
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6'
    ],

    # What does your project relate to?
    keywords='synchrotron analysis x-ray scattering tomography ',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_namespace_packages(exclude=['docs', 'tests*']),

    package_dir={},

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    # py_modules=["__init__"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=install_requires,  # 'astroscrappy' removed pending windows issue

    setup_requires=[],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,tests]
    extras_require={
        # 'dev': ['check-manifest'],
        'tests': ['pytest', 'coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={'xicam.SAXS': []},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[#('lib/python2.7/site-packages/gui', glob.glob('gui/*')),
    #            ('lib/python2.7/site-packages/yaml/tomography',glob.glob('yaml/tomography/*'))],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'databroker.ingestors': [
            'application/x-edf = xicam.SAXS.ingestors.edf_ingestor:edf_ingestor'
            # FIXME (below conflicts with ingest_nxXPCS and potentially other h5 ingestors)
            #'application/x-hdf5 = xicam.SAXS.ingestors.nxcansas:ingest_nxcanSAS'
        ],
        'xicam.plugins.GUIPlugin': [
            'SAXS = xicam.SAXS.SAXSGUIPlugin:SAXSPlugin'
        ],
        'xicam.plugins.ProcessingPlugin': [
            'CorrectFastCCDImage = xicam.SAXS.operations.correction:correct_fastccd_image',
        ],
        'xicam.plugins.OperationPlugin': [
            "fourier_autocorrelation = xicam.SAXS.calibration.fourierautocorrelation:fourier_autocorrelation",
            "naive_sdd = xicam.SAXS.calibration.naivesdd:naive_sdd",
            "array_transpose = xicam.SAXS.operations.arraytranspose:array_transpose",
            "ricker_wavelet = xicam.SAXS.calibration.cwt:ricker_wavelet",
            "simulate_calibrant = xicam.SAXS.calibration.simulatecalibrant:simulate_calibrant",
            "cake_integrate = xicam.SAXS.operations.cakeintegrate:cake_integration",
            "chi_integrate = xicam.SAXS.operations.chiintegrate:chi_integrate",
            "array_rotate = xicam.SAXS.operations.arrayrotate:array_rotate",
            "one_time_correlation = xicam.SAXS.operations.onetime:one_time_correlation",
            "two_time_correlation = xicam.SAXS.operations.twotime:two_time_correlation",
            "fit_scattering_factor = xicam.SAXS.operations.fitting:fit_scattering_factor",
            "correct_fastccd_image = xicam.SAXS.operations.correction:correct_fastccd_image",
            "chi_squared = xicam.SAXS.operations.chisquared:chi_squared",
            "astropyfit = xicam.SAXS.operations.astropyfit:AstropyQSpectraFit",
            "q_integrate = xicam.SAXS.operations.qintegrate:q_integrate",
            "fourier_correlation = xicam.SAXS.operations.fourierautocorrelator:fourier_correlation",
            "inpaint = xicam.SAXS.operations.inpaint:inpaint",
            "q_conversion_gisaxs = xicam.SAXS.operations.qconversiongisaxs:q_conversion_gisaxs",
            "q_conversion_saxs = xicam.SAXS.operations.qconversionsaxs:q_conversion_saxs",
            "x_integrate = xicam.SAXS.operations.xintegrate:x_integrate",
            "z_integrate = xicam.SAXS.operations.zintegrate:z_integrate",
            "porod_plot = xicam.SAXS.operations.porod_plot:porod_plot",
            "guinier_plot = xicam.SAXS.operations.guinier_plot:guinier_plot",
            "detector_mask = xicam.SAXS.masking.detector:detector_mask_plugin",
            "roi_mask = xicam.SAXS.masking.roi:roi_mask_plugin",
            "set_geometry = xicam.SAXS.operations.select_detector:set_geometry",
            "set_detector = xicam.SAXS.operations.select_detector:set_detector",
            "diffusion_coefficient = xicam.SAXS.operations.diffusion_coefficient:diffusion_coefficient"
        ],
        'xicam.plugins.Fittable1DModelPlugin': [
            'Gaussian1D = xicam.SAXS.models.gaussian1d:Gaussian1D'
        ],
        'xicam.plugins.SettingsPlugin': [
            'xicam.SAXS.calibration = xicam.SAXS.calibration:DeviceProfiles'
        ],
        'xicam.plugins.IntentCanvasPlugin': [
            'saxs_image_intent_canvas = xicam.SAXS.canvases:SAXSImageIntentCanvas'
        ],
        "databroker.intents": [
            "SAXSImageIntent = xicam.SAXS.intents:SAXSImageIntent",
        ],
    },

    ext_modules=[],
    include_package_data=True
)



import numpy as np

from xicam.core import msg
# from xicam.plugins.processingplugin import Input, Output, ProcessingPlugin
#
# try:
#     from csxtools.fastccd import correct_images
#     from csxtools.utils import stackmean
#
#     class CSXCorrectImage(ProcessingPlugin):
#         """Subtract backgrond and gain correct images
#         This routine subtrtacts the backgrond and corrects the images
#         for the multigain FastCCD ADC.
#         Parameters
#         ----------
#         in : array_like
#             Input array of images to correct of shape (N, y, x)  where N is the
#             number of images and x and y are the image size.
#         dark : array_like, optional
#             Input array of dark images. This should be of shape (3, y, x).
#             dark[0] is the gain 8 (most sensitive setting) dark image with
#             dark[2] being the gain 1 (least sensitive) dark image.
#         flat : array_like, optional
#             Input array for the flatfield correction. This should be of shape
#             (y, x)
#         gain : tuple, optional
#             These are the gain multiplication factors for the three different
#             gain settings
#         Returns
#         -------
#         array_like
#             Array of corrected images of shape (N, y, x)
#         """
#         # Images are expected to be of type np.uint16 (csxtools.utils.py:94)
#         bitmasked_images = Input(description="", type=np.ndarray, visible=False)
#         # Darks are expected to be of type np.float32 (csxtools.fastccd.images.py:43)
#         dark_images = Input(description="", type=np.ndarray, visible=False)
#         # Flats are expected to be of type np.float32 (csxtools.fastccd.images.py:47)
#         flat_field = Input(description="", type=np.ndarray, default=None, visible=False)
#         gain = Input(description="", type=tuple, visible=False)
#
#         corrected_images = Output(description="", type=np.ndarray)
#
#         def evaluate(self):
#             data = np.asarray(self.bitmasked_images.value.astype(np.uint16))
#             kwargs = dict()
#             reduced_dark_image = self.dark_images.value
#             if reduced_dark_image is not None:
#                 reduced_dark_image = stackmean(np.asarray(self.dark_images.value, dtype=np.float32))
#                 darks = np.array([reduced_dark_image,
#                                   np.zeros(shape=reduced_dark_image.shape),
#                                   np.zeros(shape=reduced_dark_image.shape)],
#                                  dtype=np.float32)
#                 kwargs['dark'] = darks
#             if self.flat_field.value is not None:
#                 kwargs['flat'] = self.flat_field.value.astype(np.float32)
#             if self.gain.value is not None:
#                 kwargs['gain'] = self.gain.value
#             self.corrected_images.value = correct_images(data, **kwargs)
#
# except ImportError:
#     msg.logMessage("Cannot import csxtools; will not be tested against.")
#
#
# if __name__ == "__main__":
#     shape = (5, 5)
#     dtype = np.float32
#     low = np.zeros(shape=shape, dtype=dtype)
#     high = np.ones(shape=shape, dtype=dtype) * 100
#     raw_images = np.array([low, high], dtype=np.uint16)
#     dark_image1 = np.zeros(shape=shape)
#     dark_image1[0] = 1
#     dark_image2 = np.zeros(shape=shape)
#     dark_image2[:, 3] = 1
#     dark_images = np.array([dark_image1, dark_image2], dtype=dtype)
#     print(f"dark images shape: {dark_images.shape}")
#     rr = stackmean(dark_images)
#     print(type(rr))
#     print(f"low:\n{low}")
#     print()
#     print(f"high:\n{high}")
#     print()
#     print(f"darks:\n{dark_images}")
#     print()
#     print("reduced dark image")
#     print(rr)
#     print()
#     print("corrected")
#
#     darks = np.array([
#         rr,
#         np.zeros(shape=rr.shape),
#         np.zeros(shape=rr.shape), ],
#         dtype=dtype)
#     corrected = correct_images(images=raw_images, dark=darks)
#     print(corrected)
#
#     print()
#     print('-' * 80)


### Module example (ArrayRotate from Ron)

# from xicam.plugins import ProcessingPlugin, Input, Output, InOut
# import numpy as np


# class ArrayRotate(ProcessingPlugin):
#     data = InOut(description='Array of two or more dimensions.', type=np.ndarray)
#     k = Input(description='Number of times the array is rotated by 90 degrees.', type=int, default=1)
#     axes = Input(description='The array is rotated in the plane defined by the axes. Axes must be different.',
#                  type=np.ndarray, default=(0, 1))

#     def evaluate(self):
#         self.data.value = np.rot90(self.data.value, self.k.value, self.axes.value)

### Module from QZ starts here:

from xicam.plugins import ProcessingPlugin, Input, Output, InOut
from scipy import ndimage
import numpy as np


class CoM(ProcessingPlugin):
    data = Input(description='Array of two or more dimensions.', type=np.ndarray)
    mask = Input(description='Array (same size as image) with 1 for masked pixels, and 0 for valid pixels',
                 type=np.ndarray)
    x_min = Input(description='X pixel index, bottom left.', type=int, default=1)
    y_min = Input(description='Y pixel index, bottom left.', type=int, default=1)

    x_max = Input(description='X pixel index, top right.', type=int, default=1000)
    y_max = Input(description='Y pixel index, top right.', type=int, default=1000)

    x_cen = InOut(description='X pixel index, center of mass', type=float)
    y_cen = InOut(description='Y pixel index, center of mass', type=float)

    def evaluate(self):
        data = np.flipud(self.data.value)
        mask = np.flipud(self.mask.value)
        data = data * np.logical_not(mask)
        (self.y_cen.value, self.x_cen.value) = ndimage.measurements.center_of_mass(
            data[self.y_min.value:self.y_max.value,
            self.x_min.value:self.x_max.value])
        self.x_cen.value = self.x_cen.value + self.x_min.value
        self.y_cen.value = self.y_cen.value + self.y_min.value



from xicam.plugins import ProcessingPlugin, Input, Output
import numpy as np
from xicam.plugins.hint import VerticalROI

class VerticalCutPlugin(ProcessingPlugin):
    data = Input(description='Frame image data', type=np.ndarray)
    qz = Input(description='qz coordinate corresponding to data', type=np.ndarray)

    mask = Input(description='Frame image data', type=np.ndarray, default=None)

    # Make qz range a single parameter, type = tuple
    qzminimum = Input(description='qz minimum limit', type=int)
    qzmaximum = Input(description='qz maximum limit', type=int)

    cut = Output(description='mask (1 is masked) with dimension of data', type=np.ndarray)

    hints = [VerticalROI(qzminimum, qzmaximum)]

    def evaluate(self):
        if self.mask.value is not None:
            self.cut.value = np.logical_or(self.mask.value, self.qz < self.qzminimum.value,
                                           self.qz > self.qzmaximum.value)
        else:
            self.cut.value = np.logical_or(self.qz.value < self.qzminimum.value,
                                           self.qz.value > self.qzmaximum.value)



import numpy as np
from numba import njit, prange
from typing import Iterable, Tuple
from xicam.plugins.operationplugin import operation, display_name, output_names, describe_input, \
    describe_output, categories, visible
import pyqtgraph as pg
from ..utils import get_label_array
from xarray import DataArray


def correct(array, flats, bkg, gain_map=(1, 2, 4, 8), clip=False):
    return _correct(np.asarray(array, dtype=np.uint16),
                    np.asarray(flats, dtype=np.uint16),
                    np.asarray(bkg, dtype=np.uint16),
                    clip=clip)


@njit(parallel=True)
def _correct(array, flats, bkg, gain_map=(1, 2, 4, 8), clip=False):
    # 16-bit unsigned
    # image: bits 0 - 12
    # bad:   bit 13
    # gain:  bits 14-15
    # gain1: 0b11 (3)
    # gain2: 0b10 (2)
    # gain8: 0b00 (0)
    masked_flat = 0x1FFF & flats
    masked_dark = 0x1FFF & bkg
    out = np.empty_like(array, dtype=np.int32)

    for i in prange(array.shape[0]):
        for j in prange(array.shape[1]):
            for k in prange(array.shape[2]):
                val = array[i, j, k]

                intensity = 0x1FFF & val
                bad_flag = 0x1 & (val >> 13)
                gain = gain_map[0x3 & (val >> 14)]

                bkg_gain = gain_map[0x3 & (bkg[j, k] >> 14)]

                out[i, j, k] = masked_flat[j, k] * gain * intensity

                if clip:
                    if intensity < masked_dark[j, k]:
                        out[i, j, k] = bkg_gain * masked_dark[j, k]

                if bad_flag:
                    out[i, j, k] = 0
                else:
                    out[i, j, k] -= bkg_gain * masked_dark[j, k]

    return out


def reduce_data(images, slices):
    trimmed_images = []
    for i in range(len(images)):
        image = np.asarray(images[i])
        image = image[slices]
        trimmed_images.append(image)
        del image

    return DataArray(np.asarray(trimmed_images))


@operation
@display_name('FastCCD Correction')
@describe_input('images', '3-dimensional input array containing 2-dimensional images')
@describe_input('flats', '2-dimensional input array containing flat image data')
@describe_input('darks', '3-dimensional input array containing 2-dimensional dark image data')
@describe_input('gains', 'n-dimensional input array containing gain values')
@output_names('images', 'labels')
@describe_output('images', 'Corrected fastccd image data')
@categories(('Scattering', 'Calibration'))
@visible('images', False)
@visible('flats', False)
@visible('darks', False)
def correct_fastccd_image(images: np.ndarray,
                          flats: np.ndarray = None,
                          darks: np.ndarray = None,
                          gains: tuple = (1, 2, 4, 8),
                          clip_below_dark: bool = False,
                          rois: Iterable[pg.ROI] = None,
                          image_item: pg.ImageItem = None,
                          ) -> Tuple[np.ndarray, np.ndarray]:
    # TODO: is this pulling from the correct dark? we need of mapping of gain index to dark array?

    if images.ndim not in [2, 3]:
        raise ValueError(f"\"images\" expects a 2- or 3-dimensional image array; shape = \"{images.shape}\"")

    flats = flats
    if flats is None:
        flats = np.ones_like(images[0])
    elif flats.ndim != 2:
        raise ValueError(f"\"flats\" should be 2-dimensional; shape = \"{flats.shape}\"")

    darks = darks
    if darks is None:
        darks = np.zeros_like(images[0])
    elif darks.ndim == 3:
        darks = np.sum(darks, axis=0) / darks.shape[0]
    elif darks.ndim == 2:
        pass  # darks is already a single frame

    labels = np.zeros((1, *images.shape[-2:]))
    if rois and image_item:
        labels = get_label_array(images, rois=rois, image_item=image_item)

        # Trim the image based on labels, and resolve to memory
        si, se = np.where(np.flipud(labels))
        # images = np.asarray(images[:, si.min():si.max() + 1, se.min():se.max() + 1])
        slices = slice(si.min(), si.max() + 1), slice(se.min(), se.max() + 1)
        template = DataArray(np.empty((images.shape[0], si.max() - si.min() + 1, se.max() - se.min() + 1))).chunk(
            images._variable.data.chunksize)
        images = images.map_blocks(reduce_data, template=template, kwargs=dict(slices=slices)).compute()

        flats = np.asarray(flats[si.min():si.max() + 1, se.min():se.max() + 1])
        darks = np.asarray(darks[si.min():si.max() + 1, se.min():se.max() + 1])
        labels = np.asarray(np.flipud(labels)[si.min():si.max() + 1, se.min():se.max() + 1])

    corrected_images = correct(images,
                               flats,
                               darks,
                               gains,
                               clip=clip_below_dark)
    return corrected_images, labels



from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
import numpy as np
import pyFAI
# FIXME: where is this package installed from?
import hipies


@operation
@display_name('Remesh')
@describe_input('data', 'Detector image')
@describe_input('geometry', 'pyFAI Geometry')
@describe_input('alpha_in', 'GISAXS incidence angle')
@describe_input('out_range', 'coordinates of output image')
@describe_input('resolution', 'resolution of output image')
@describe_input('coord_sys', 'Choice of coordinate system for output image')
@output_names('I', 'x', 'y', 'geometry')
@describe_output('I', 'remapped image')
@describe_output('x', 'X-coordinates output image')
@describe_output('y', 'Y-coordinates output image')
@describe_output('geometry', 'pyFAI geometry')

def image_remap(data: np.ndarray,
                geometry: pyFAI.geometry.Geometry,
                alpha_in: float,
                out_range: list=None,
                resolution: list=None,
                coord_sys: str='qp_qz') -> np.ndarray:

    I, x, y = hipies.remesh(data, geometry, alpha_in, out_range = out_range, 
                         res = resolution, coord_sys = coord_sys)
    dist = geometry._dist
    centerX = np.unravel_index(x.abs().argmin(), x.shape)[1]
    centerY = np.unravel_index(y.abs().argmin(), y.shape)[0]
    pixel = [geometry.get_pixel1(), geometry.get_pixel2()]
    center = [centerX * pixel[0], centerY * pixel[1]]
    geometry.setFit2D(geometry._dist, center[0], center[1])

    return I, x, y, geometry




import time

import event_model
import h5py
from dask import array as da
from pathlib import Path
from xarray import DataArray
import mimetypes

mimetypes.add_type('application/x-hdf5', '.hdf5')
I_PROJECTION_KEY = 'sasentry/sasdata/I'
QX_PROJECTION_KEY = 'sasentry/sasdata/Qx'
QY_PROJECTION_KEY = 'sasentry/sasdata/Qy'


def ingest_nxcanSAS(paths):
    assert len(paths) == 1
    path = paths[0]

    h5 = h5py.File(path, 'r')

    data = h5['sasentry']['sasdata']['I']
    Qx = h5['sasentry']['sasdata']['Qx'][()]
    Qy = h5['sasentry']['sasdata']['Qy'][()]
    axes = h5['sasentry']['sasdata'].attrs['I_axes']

    xarray = DataArray(data, dims=axes.split(','))
    dask_data = da.from_array(xarray)

    projections = [{'name': 'NXcanSAS',
                    'version': '0.1.0',
                    'projection':
                        {I_PROJECTION_KEY: {'type': 'linked',
                                                'stream': 'primary',
                                                'location': 'event',
                                                'field': 'I'},
                         QX_PROJECTION_KEY: {'type': 'linked',
                                                 'stream': 'primary',
                                                 'location': 'event',
                                                 'field': 'Qx'},
                         QY_PROJECTION_KEY: {'type': 'linked',
                                                 'stream': 'primary',
                                                 'location': 'event',
                                                 'field': 'Qy'},
                         # 'sasentry/sasdata/': {'type': 'linked',
                         #                          'stream': 'primary',
                         #                          'location': 'event',
                         #                          'field': 'g2'},
                         }

                    }]
    # Compose run start
    run_bundle = event_model.compose_run()  # type: event_model.ComposeRunBundle
    start_doc = run_bundle.start_doc
    start_doc["sample_name"] = Path(paths[0]).resolve().stem
    start_doc["projections"] = projections
    yield 'start', start_doc

    # Compose descriptor
    source = 'NXcanSAS'
    frame_data_keys = {'I': {'source': source,
                             'dtype': 'number',
                             'dims': xarray.dims,
                             # 'coords': [energy, sample_y, sample_x],
                             'shape': data.shape},
                       'Qx': {'source': source,
                              'dtype': 'number',
                              'shape': Qx.shape, },
                       'Qy': {'source': source,
                              'dtype': 'number',
                              'shape': Qy.shape, }
                       }

    frame_stream_name = 'primary'
    frame_stream_bundle = run_bundle.compose_descriptor(data_keys=frame_data_keys,
                                                        name=frame_stream_name,
                                                        # configuration=_metadata(path)
                                                        )
    yield 'descriptor', frame_stream_bundle.descriptor_doc

    # NOTE: Resource document may be meaningful in the future. For transient access it is not useful
    # # Compose resource
    # resource = run_bundle.compose_resource(root=Path(path).root, resource_path=path, spec='NCEM_DM', resource_kwargs={})
    # yield 'resource', resource.resource_doc

    # Compose datum_page
    # z_indices, t_indices = zip(*itertools.product(z_indices, t_indices))
    # datum_page_doc = resource.compose_datum_page(datum_kwargs={'index_z': list(z_indices), 'index_t': list(t_indices)})
    # datum_ids = datum_page_doc['datum_id']
    # yield 'datum_page', datum_page_doc

    yield 'event', frame_stream_bundle.compose_event(data={'I': dask_data,
                                                           'Qx': Qx,
                                                           'Qy': Qy},
                                                     timestamps={'I': time.time(),
                                                                 'Qx': time.time(),
                                                                 'Qy': time.time()})

    yield 'stop', run_bundle.compose_stop()


if __name__ == "__main__":
    path = 'test21.hdf5'

    docs = list(ingest_nxcanSAS([path]))

    print(docs)



import inspect
import pytest

import numpy as np

from xicam.core.execution.daskexecutor import DaskExecutor
from xicam.core.execution.workflow import Workflow
from xicam.SAXS.operations.correction import correct_fastccd_image

# Try to compare against the csxtools implementation...
# Differences:
#     * csxtools returns a float array (see fastccd.h:data_t)
#         - values below the dark array are negative
#         - bad pixels set to NaN
#     * we return a uint16 array
#         - values below the dark array become 0 (they do NOT wrap around the uint16)
#         - bad pixels set to 0
CSX_TOOLS = False
try:
    from . import CSXCorrectImage
    CSX_TOOLS = True
except ImportError:
    CSXCorrectImage = None

# Disable testing against csxtools here
TEST_CSX_TOOLS = CSX_TOOLS and True


# TODO: test endianess
# TODO: test non-uint16 inputs?

# Used for testing purposes
executor = DaskExecutor()

@pytest.fixture
def op() -> correct_fastccd_image:
    op = correct_fastccd_image()
    op.filled_values['images'] = np.arange(24, dtype=np.uint16).reshape((4, 3, 2))
    return op

@pytest.fixture
def old():
    if TEST_CSX_TOOLS:
        op = CSXCorrectImage()
        op.filled_values['bitmasked_images'] = np.arange(24, dtype=np.uint16).reshape((4, 3, 2))
        return op
    else:
        return None

def get_default(operation, param_name: str):
    # TODO: extend operation api with ability to do this
    parameters = inspect.signature(operation._func).parameters
    param = parameters.get(param_name, None)
    if param:
        return param.default
    return None


class TestCorrectImage:

    def test_default(self, op, old):
        w = Workflow(operations=[op])
        result = w.execute_synchronous(executor=executor)
        corrected_images = result[0]['corrected_images']
        gains = get_default(op, 'gains')
        assert np.array_equal(corrected_images, op.filled_values['images'] * gains[0])
        if TEST_CSX_TOOLS:
            w.clear_operations()
            w.add_operation(old)
            result = w.execute_synchronous(executor=executor)
            assert np.array_equal(result[0]['corrected_images'], corrected_images)

    def test_simple_darks(self, op, old):
        op.darks.value = np.ones(shape=(1, *op.images.value.shape[1:]))  # darks of 1's
        op.evaluate()
        truth_value = np.array(op.images.value * op.gains.value[0] - op.darks.value, dtype=np.uint16)
        truth_value[op.darks.value > op.images.value] = 0
        assert np.array_equal(op.corrected_images.value, truth_value)
        if TEST_CSX_TOOLS:
            old.dark_images.value = op.darks.value
            old.evaluate()
            # csxtools gives us negative values; our code will zero out values less than the dark values
            old.corrected_images.value = np.where(old.corrected_images.value < 0, 0, old.corrected_images.value)
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_multiple_darks(self, op, old):
        # Should average to dark array of 2's
        op.darks.value = np.array([np.ones(op.images.value.shape[1:]),
                                   np.ones(op.images.value.shape[1:]) * 3])
        op.evaluate()
        truth_value = np.array(op.images.value * op.gains.value[0] - 2, dtype=np.uint16)
        truth_value[2 > op.images.value] = 0
        assert np.array_equal(op.corrected_images.value, truth_value)
        if TEST_CSX_TOOLS:
            old.dark_images.value = op.darks.value
            old.evaluate()
            # csxtools gives us negative values; our code will zero out values less than the dark values
            old.corrected_images.value = np.where(old.corrected_images.value < 0, 0, old.corrected_images.value)
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_simple_flats(self, op, old):
        op.flats.value = np.ones(shape=op.images.value.shape[1:]) * 2
        op.evaluate()
        truth_value = np.array(op.images.value * op.gains.value[0] * op.flats.value, dtype=np.uint16)
        assert np.array_equal(op.corrected_images.value, truth_value)
        if TEST_CSX_TOOLS:
            old.flat_field.value = op.flats.value
            old.evaluate()
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_gain_1(self, op, old):
        # gain1: 0b11 -> 1100 -> 0xC
        op.images.value = np.bitwise_or(0xC000, op.images.value)
        op.evaluate()
        truth_value = np.bitwise_and(0x1FFF, op.images.value * op.gains.value[2])
        assert np.array_equal(op.corrected_images.value, truth_value)
        if TEST_CSX_TOOLS:
            old.bitmasked_images.value = op.images.value
            old.evaluate()
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_gain_2(self, op, old):
        # gain2: 0b10 -> 1000 -> 0x8
        op.images.value = np.bitwise_or(0x8000, op.images.value)
        op.evaluate()
        truth_value = np.bitwise_and(0x1FFF, op.images.value * op.gains.value[1])
        assert np.array_equal(op.corrected_images.value, truth_value)
        if TEST_CSX_TOOLS:
            old.bitmasked_images.value = op.images.value
            old.evaluate()
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_gain_8(self, op, old):
        # gain8: 0b00 -> 0000 -> 0x0
        op.gains.value = (2, 1, 1)  # explicitly set our gain8 channel to 2
        op.images.value = np.bitwise_or(0x0000, op.images.value)
        op.evaluate()
        assert np.array_equal(op.corrected_images.value, op.images.value * op.gains.value[0])
        if TEST_CSX_TOOLS:
            old.bitmasked_images.value = op.images.value
            old.gain.value = op.gains.value
            old.evaluate()
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    def test_bad_pixel(self, op, old):
        # bad pixel: -> 0010 -> 0x2
        op.images.value[:, -1, -1] = np.bitwise_or(0x2000, op.images.value[:, -1, -1])
        op.evaluate()
        truth_value = np.bitwise_and(0x1FFF, op.images.value * op.gains.value[0])
        truth_value[:, -1, -1] = np.nan
        assert np.array_equal(op.corrected_images.value, truth_value)
        
        # csxtools gives back NaN for bad pixels; we give back 0's
        if TEST_CSX_TOOLS:
            old.bitmasked_images.value = op.images.value
            old.evaluate()
            # Convert the NaN's to 0
            old.corrected_images.value = np.where(np.isnan(old.corrected_images.value), 0, old.corrected_images.value)
            assert np.array_equal(old.corrected_images.value, op.corrected_images.value)

    # def test_no_input_images(self, op, old):
    #     op.images.value = None
    #     with pytest.raises(TypeError):
    #         op.evaluate()
    #
    # def test_empty_input_images(self, op, old):
    #     op.images.value = []
    #     with pytest.raises(TypeError):
    #         op.evaluate()

    def test_bad_input_images_shape(self, op, old):
        op.images.value = np.ones(shape=(10, 10))
        with pytest.raises(ValueError):
            op.evaluate()

    def test_bad_flat_shape(self, op, old):
        op.flats.value = np.ones(shape=(10, 10, 10))
        with pytest.raises(ValueError):
            op.evaluate()

    # def test_bad_flat_type(self, op, old):
    #     op.flats.value = []
    #     with pytest.raises(TypeError):
    #         op.evaluate()

    def test_bad_darks_shape(self, op, old):
        op.darks.value = np.zeros(shape=(10, 10))
        with pytest.raises(ValueError):
            op.evaluate()

    # def test_bad_darks_type(self, op, old):
    #     op.darks.value = []
    #     with pytest.raises(TypeError):
    #         op.evaluate()



from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import numpy as np


@operation
@display_name("Detector Mask")
@describe_input('azimuthal_integrator', "The `AzimuthalIntegrator` to including a detector instance describing the camera hardware.")
@describe_output("mask", "The calculated mask derived from the detector's profile. (1 is masked)")
@output_names("mask")
@categories(("Scattering", "Masking"))
def detector_mask_plugin(azimuthal_integrator: AzimuthalIntegrator = None, mask:np.ndarray = None) -> np.ndarray:
    if azimuthal_integrator and azimuthal_integrator.detector:
        if mask is None:
            mask = np.zeros(azimuthal_integrator.detector.shape)
        mask = np.logical_or(azimuthal_integrator.detector.calc_mask(), mask)
        return mask



import os
import entrypoints
import numpy as np


def test_edf_ingestor(tmp_path):
    from fabio.edfimage import EdfImage
    # test data
    data = np.random.random((1000, 1000))
    # write data to test edf
    edf_path = os.path.join(tmp_path, "test.edf")
    print("edf_path:", edf_path)
    EdfImage(data).write(edf_path)

    # get edf ingestor
    edf_ingestor = entrypoints.get_single("databroker.ingestors", "application/x-edf").load()

    # load data into catalog
    document = list(edf_ingestor([edf_path]))
    uid = document[0][1]["uid"]

    # TODO: actually do some assertions in here (assert the keys are correct in the descriptor)
    #  (assert data is image) ...


from xicam.plugins.datahandlerplugin import DataHandlerPlugin, start_doc, descriptor_doc, event_doc, stop_doc, \
    embedded_local_event_doc

import os
import fabio
import uuid
import re
import functools
from pathlib import Path




class TIFPlugin(DataHandlerPlugin):
    name = 'TIFPlugin'

    DEFAULT_EXTENTIONS = ['.tiff', '.tif']

    descriptor_keys = ['object_keys']

    def __call__(self, *args, **kwargs):
        return fabio.open(self.path).data

    def __init__(self, path):
        super(TIFPlugin, self).__init__()
        self.path = path
        self.fimg = fabio.open(path)

    @staticmethod
    @functools.lru_cache(maxsize=10, typed=False)
    def parseDataFile(path):
        md = fabio.open(path).header
        md.update({'object_keys': {'pilatus2M': ['primary']}})
        return md

    @classmethod
    def getStartDoc(cls, paths, start_uid):
        return start_doc(start_uid=start_uid, metadata={'paths': paths})



import pyqtgraph as pg
from xicam.core import msg
from xicam.core.intents import ImageIntent
from xicam.plugins.operationplugin import operation, describe_input, describe_output, visible, \
    input_names, output_names, display_name, categories, intent
import numpy as np
from skbeam.core.correlation import two_time_corr
from typing import Tuple, Iterable

from ..patches.pyFAI import AzimuthalIntegrator
from ..utils import average_q_from_labels, get_label_array


@operation
@display_name('2-time Correlation')
# @input_names('images', 'labels', 'number_of_buffers', 'number_of_levels')
@describe_input('images', 'dimensions are: (rr, cc), iterable of 2D arrays')
@describe_input('labels', 'labeled array of the same shape as the image stack;'
                          'each ROI is represented by a distinct label (i.e., integer)')
@describe_input('autoset_num_bufs', 'When enabled, automatically determine the number of buffers to be used. '
                                    'This will ignore any value in num_bufs if checked. '
                                    'Enabling this will set number of levels to 1 (if not already set to 1). '
                                    'The number of buffers selected will be automatically based on number of frames,'
                                    ' (or number of frames - 1 if there are an odd number of frames.')
@describe_input('num_bufs', 'Maximum lag step to compute in each generation of downsampling (must be even). '
                            'Check autoset_num_bufs to automatically set this for you.')
@describe_input('num_levels', 'how many generations of downsampling to perform,'
                              'i.e., the depth of the binomial tree of averaged frames default is one')
@output_names('g2', 'tau', 'qs')
@describe_output('g2', 'the normalized correlation shape is (num_rois, len(lag_steps), len(lag_steps))')
@describe_output('tau', 'the times at which the correlation was computed')
@visible('images', False)
@visible('rois', False)
@intent(ImageIntent,
        name='2-time Correlation',
        output_map={'image': 'g2', 'xvals': 'qs'},
        mixins=["AxesLabels", "XArrayView", "SliceSelector"],
        labels={"bottom": "𝜏₁", "left": "𝜏₂"})
def two_time_correlation(images: np.ndarray,
                         image_item: pg.ImageItem = None,
                         rois: Iterable[pg.ROI] = None,
                         autoset_num_bufs: bool = True,
                         num_bufs: int = 2,
                         num_levels: int = 1,
                         geometry: AzimuthalIntegrator = None) -> Tuple[np.ndarray, np.ndarray]:
    # TODO -- make composite parameter item widget to allow default (all frames) or enter value
    num_frames = len(images)

    # Auto-select-buffers will override number of levels to be 1,
    # which requires that the number of buffers == number of frames
    # (unless odd; then use number of frames - 1)
    if autoset_num_bufs:
        num_levels = 1
        num_frames = num_frames if num_frames % 2 == 0 else num_frames - 1
        num_bufs = num_frames

    labels = get_label_array(images, rois=rois, image_item=image_item)
    if labels.max() == 0:
        msg.notifyMessage("Please add an ROI over which to calculate one-time correlation.")
        raise ValueError("Please add an ROI over which to calculate one-time correlation.")

    corr = two_time_corr(labels.astype(np.int_),
                         np.asarray(images),
                         num_frames,
                         num_bufs,
                         num_levels)
    g2 = corr.g2
    lag_steps = corr.lag_steps

    # Calculate avg qs from label array for first dimension on returned g2 (so slice selector shows qs for indexing)
    qs = None
    if geometry is not None:
        qs = np.asarray(average_q_from_labels(labels, geometry, scattering_mode='transmission'))
        # qs = np.asarray([f"q={q:.3f}" for q in qs])  # FIXME: why can't we return a python list for the catalog?
        # File "xi-cam/xicam/core/execution/workflow.py", line 886, in project_intents
        #     kwargs[intent_kwarg_name] = getattr(run_catalog, operation_id).to_dask()[output_name]
        # ...
        # File "site-packages/xarray/core/dataarray.py", line 126, in _infer_coords_and_dims
        #     raise ValueError(
        #  ValueError: different number of dimensions on data and dims: 2 vs 1

    # Rotate image plane 90 degrees
    g2 = np.rot90(g2, axes=(-2, -1))

    num_labels = g2.shape[0]  # first dimension represents labels
    if not qs:
        qs = np.array(list(range(1, num_labels + 1)))
    return g2, lag_steps, qs



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import yaml
from collections import OrderedDict
from pyqtgraph.parametertree import Parameter

cfg = 'config.yml'

# Fitting related
_fixed = {'name': 'Fixed', 'value': True, 'type': 'bool'}
_bounds = [{'name': 'Lower', 'value': '-\u221E'}, {'name': 'Upper', 'value': '\u221E'}]
_bounded = {'name': 'Bounded', 'value': False, 'type': 'bool',
            'children': _bounds, 'visible': False, 'enabled': False}


# resolve metaclass conflict
class XiMetaParam(type(yaml.YAMLObject), type(Parameter)):
    pass


class XiCamParameter(yaml.YAMLObject, Parameter, metaclass=XiMetaParam):
    yaml_tag = u'!YMLParameter'

    def __init__(self, name, description, value, units, **kwags):
        Parameter.__init__(self, name=name, value=value, \
                           title=description, children=[_fixed, _bounded])

    def __repr__(self):
        return "%s (%r, value=%r)" % \
               (self.__class__.__name__, self.name(), self.value())

    @classmethod
    def from_yaml(cls, loader, node):
        opts = loader.construct_mapping(node)
        return cls(**opts)


def load_models():
    with open(cfg) as fp:
        yml = yaml.load(fp)

    model_tree = OrderedDict()
    for key, val in yml.items():
        models = OrderedDict()
        for name, params in val.items():
            _params = [p['param'] for p in params]
            models[name] = {'params': _params}
        model_tree[key] = models
    return model_tree


if __name__ == '__main__':
    models = load_models()
    model = models['Cylinder Functions']['Barbell']['params']
    print(model)



from xicam.plugins import ProcessingPlugin, Input, Output, InOut
import numpy as np
import astroscrappy


class ZingerMaskPlugin(ProcessingPlugin):
    data = Input(description='Frame image data',
                 type=np.ndarray)
    mask = InOut(description='Mask array (1 is masked).',
                 type=np.ndarray)

    def evaluate(self):
        self.mask.value = np.logical_or(self.mask.value,
                                        astroscrappy.detect_cosmics(self.data.value, self.mask.value)[0])
